/*********** blocks.c *****************
	these routines perform the block operations.
	usually they work only in the active window.
	also goto_line is defined here.
	all routines are not system dependent!
***************************************/

#include <string.h>

#include "types.h"
#include "sysdep.h"
#include "blocks.h"
#include "files.h"
#include "header.h"
#include "egraphic.h"
#include "key.h"
#include "menu.h"

blocktyp block,notedbl;
int copied=0;

void block_start (void)
/********* block_start
	define block start.
*********/
{	int slot;
	filetyp *ft;
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	if (block.slot!=slot) block.end=ft->fend;
	find_pos(ft->cursor.line,ft->cursor.x,ft->fend,&(ft->cursor.adr),
		ft->tabs);
	block.start=ft->cursor.adr;
	if (block.slot!=slot) fredraw(block.slot);
	block.slot=slot;
	if (!block.end) block.end=ft->fend;
	if (block.start>=block.end) block.slot=-1;
	fredraw(slot);
}

void block_end (void)
/********* block_end
	define block end.
*********/
{	int slot;
	filetyp *ft;
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	if (block.slot!=slot) block.start=ft->start;
	find_pos(ft->cursor.line,ft->cursor.x,ft->fend,&(ft->cursor.adr),
		ft->tabs);
	block.end=ft->cursor.adr;
	if (block.slot!=slot) fredraw(block.slot);
	block.slot=slot;
	if (!block.start) block.start=ft->start;
	if (block.start>=block.end) block.slot=-1;
	fredraw(slot);
}

void block_hide (void)
/********* block_hide
	undefine the block.
**********/
{	int slot;
	block.start=block.end=0;
	slot=block.slot;
	block.slot=-1;
	if (slot>=0) fredraw(slot);	
}

void block_delete (void)
/********** block_delete
	delete the block.
**********/
{	filetyp *ft;
	long size;
	int slot;
	if (block.slot<0 || !block.start || !block.end) return;
	ft=files+block.slot;
	size=block.end-block.start;
	to_undo(block.start,size,block.slot);
	fdelete(block.slot,block.start,size);
	ft->flags.updated=1;
	size=((ft->end-ft->fend-RESERVE)/2)*2;
	if (size>0) memory(block.slot,-size);
	slot=block.slot; block.slot=-1; block.start=block.end=0;
	fredraw(slot);
}

void block_copy (void)
/********** block_copy 
	copy a block into the activ window at cursor position 
**********/
{	filetyp *ft;
	long size;
	int slot;
	char *p;
	/* is there any block? */
	if (block.slot<0 || !block.start || !block.end) return;
	/* determine activ window */
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	/* block size */
	size=block.end-block.start;
	if (size<=0) return;
	/* be sure to have the right ft->cursor.adr (cursor addres) */
	find_pos(ft->cursor.line,ft->cursor.x,ft->fend,&(ft->cursor.adr),
		ft->tabs);
	p=ft->cursor.adr;
	if (block.start<p && block.end>p)
	{	error("Cannot copy block into itself!");
		return;
	}
	/* close the gap in the block file */
	close_gap(block.slot);
	/* make space for block, if possible! */
	if (!finsert(slot,p,size)) return;
	/* copy the block into this space */
	memmove(p,block.start,size);
	ft->flags.updated=1; /* file is updated */
	if (copied>=0) /* flag to prevent double redraw */
		fredraw(slot);
	/* note old block */
	notedbl.slot=slot;
	notedbl.start=p;
	notedbl.end=p+size;
	copied=1; /* flag for success */
}

void block_move (void)
/************ block_move
	move the block to the cursor position.
************/
{	blocktyp bl;
	copied=-1; block_copy();
	if (copied>0) 
	{	bl=block;
		block=notedbl;
		fdelete(bl.slot,bl.start,bl.end-bl.start);
		files[bl.slot].flags.updated=1;
		fredraw(block.slot);
	}
	copied=0;
}

int get_line (void)
/********* get_line
	determine line numer in active window.
	return -1, if there is no active window.
*********/
{	int n,slot;
	filetyp *ft;
	char *p,*pn;
	slot=get_top();
	if (slot<0) return -1;
	ft=files+slot;
	n=0;
	p=ft->start;
	while (p<ft->cursor.line)
	{	pn=next_line(p,ft);
		if (pn==p) return -1;
		p=pn; n++;
	}
	return n+1;
}

void goto_line (int nn)
/********* goto_line
	goto line numer nn in active window.
*********/
{	int n,slot,i;
	filetyp *ft;
	char *p,*pn;
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	p=ft->start;
	n=1;
	while (n<nn)
	{	pn=next_line(p,ft);
		if (pn==p) break;
		p=pn; n++;
	}
	ft->cursor.line=p; ft->cursor.x=0;
	for (i=0; i<5; i++) p=prev_line(p,ft);
	ft->topl=p;
	fredraw(slot);
	edit_line(0);
}

void block_save (void)
{	static char path[256]="";
	if (block.slot<0)
	{	error("No block defined!");
		return;
	}
	select_file("Save block to : ",path);
	if (params.dup) make_dup(path);
	close_gap(block.slot);
	if (path[0]) 
		memsave(path,block.start,block.end-block.start);
	refresh_top();
}

void block_load (void)
/*********
	load a block into memory
**********/
{	size_t size;
	int slot,mode;
	static char path[256]="";
	char name[16],*p;
	slot=get_top();
	if (slot<0) return;
	select_file("Load block from : ",path);
	if (!search_file(path,name,&size,&mode))
	{	error("File does not exist");
		return;
	}
	p=files[slot].cursor.adr;
	if (finsert(slot,p,size))
		if (memload(path,p,size))
		{	block_hide();
			block.start=p;
			block.end=block.start+size;
			block.slot=slot;
			files[block.slot].flags.updated=1;
		}
		else
		{	fdelete(slot,p,size);
		}
	fredraw(slot);
}

void left_block (void)
/*********** left_block
	shift block one TAB to the left.
***********/
{	char *line,*l;
	filetyp *ft;
	if (block.slot<0) return;
	ft=files+block.slot;
	update_gap(block.start,block.slot);
	line=prev_line(block.start+2,ft);
	while (line<block.end)
	{	if (*line==9) fdelete(block.slot,line,1);
		l=next_line(line,ft);
		if (line==l) break;
		else line=l;
	}
	fredraw(block.slot);
	ft->flags.updated=1;
}

void right_block (void)
/******** right_block
	insert one TAB into all lines of block.
********/
{	char *line,*l;
	filetyp *ft;
	if (block.slot<0) return;
	ft=files+block.slot;
	update_gap(block.start,block.slot);
	line=prev_line(block.start+2,ft);
	while (line<block.end)
	{	finsert(block.slot,line,1);
		*line=9;
		l=next_line(line,ft);
		if (line==l) break;
		else line=l;
	}
	fredraw(block.slot);
	ft->flags.updated=1;
}

void line_block (void)
{	int slot;
	filetyp *ft;
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	block.slot=slot;
	block.start=ft->cursor.line;
	block.end=next_line(ft->cursor.line,ft);
	if (block.end==block.start) block.end=ft->fend;
	fredraw(slot);
}

void reformat (int slot, filetyp *ft, char *line, int flag)
/******** reformat
	retformat a line.
	end, if block.end is reached.
*********/
{	char *p,*br,*p1,*p2;
	int length;
	start:
	/* delete starting white space : */
	if (flag)
	{	p=line;
		while ((*p==9)||(*p==' ')) p++;
		fdelete(slot,line,p-line);
	}
	/* determine position, which exceeds linelength */
	p=line; length=0;
	while (length<ft->linelength)
	{	if (*p==9)  length=(length/ft->tabs+1)*ft->tabs;
		else if (*p==10) /* join lines */
			{	if (p+1>=block.end) return;
				p1=p-1; 
				while ((p1>=line)&&((*p1==' ')||(*p1==9))) p1--;
				p2=p+1;
				while ((p2<block.end)&&	
					((*p2==' ')||(*p2==9))) p2++;
				if (p2>=block.end) return;
				fdelete(slot,p1+2,p2-(p1+2));
				*(p1+1)=' ';
				goto start;
			}
		else length++; 
		p++;
		if (p>=block.end) return;
	}
	br=p;
	while (br>=line)
	{	if ((*br!=9)&&(*br!=' ')&&(*br!=13)) br--; 
		else break;
	}
	if (br<line)
	{	p=line;
		while ((*p!=13)&&(*p!=' ')&&(*p!=9)) p++;
		br=p-1;
	}
	else while ((*br==' ')||(*br==9)||(*br==13)) br--;
	/* split the line */
	finsert(slot,br+1,1);
	*(br+1)=10;
	files[slot].flags.updated=1;
}

void block_format (void)
/******** block_format
	reformats a block to lines of maximal length ft->linenength.
	if there is no block, it creates one in the active window,
	as long as there are lines, which are not empty and dont start
	with white space.
**********/
{	char *line,*l;
	filetyp *ft;
	int slot,blockdef=0;
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	if (block.slot<0)
	{	block.start=line=ft->cursor.line;
		line=next_line(line,ft);
		block.end=line;
		if (block.start==block.end) { block.end=ft->fend; }
		else
		while ((*line!=' ')&&(*line!=9)&&(*line!=10))
		{	l=next_line(line,ft);
			if (l==line)
			{	block.end=fend(ft); break;
			}
			block.end=line=l;
		}
		block.slot=slot; blockdef=1;
	}
	if (slot!=block.slot) return;
	line=prev_line(block.start+2,ft);
	while (line<block.end)
	{	reformat(slot,ft,line,line>block.start);
		l=next_line(line,ft);
		if (l==line) break;
		else line=l;
	}
	if (blockdef)  { block.slot=-1; block.start=block.end=0; }
	/* update cursor position */
	if (ft->cursor.line>block.start && ft->cursor.line<block.end)
	{	ft->cursor.adr=
			ft->cursor.line=prev_line(ft->cursor.line,ft);
		ft->cursor.x=0;
	}
	fredraw(slot);
	files[slot].flags.updated=1;
}
